# sage_setup: distribution = sagemath-categories
"""
Polynomial Compilers

AUTHORS:

- Tom Boothby, initial design & implementation
- Robert Bradshaw, bug fixes / suggested & assisted with significant design improvements
"""

################################################################################
#       Copyright (C) 2007 Tom Boothby <boothby@u.washington.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
################################################################################

from sage.misc.binary_tree cimport BinaryTree


cdef class CompiledPolynomialFunction:
    """
    Build a reasonably optimized directed acyclic graph representation
    for a given polynomial.  A ``CompiledPolynomialFunction`` is callable from
    python, though it is a little faster to call the eval function from
    pyrex.

    This class is not intended to be called by a user, rather, it is
    intended to improve the performance of immutable polynomial objects.

    .. TODO::

        - Recursive calling
        - Faster casting of coefficients / argument
        - Multivariate polynomials
        - Cython implementation of Pippenger's Algorithm that doesn't
          depend heavily upon dicts.
        - Computation of parameter sequence suggested by Pippenger
        - Univariate exponentiation can use Brauer's method to improve
          extremely sparse polynomials of very high degree
    """

    def __init__(self, coeffs, algorithm='binary'):
        """
        Compiles a polynomial into an evaluation DAG representation which
        is at least as optimal as using Horner's Rule.  For polynomials
        which have a relatively large number of zero coefficients, the
        improvement over Horner's Rule grows significantly.

        Here is a rough description of the algorithm.  Actual
        implementation differs slightly, for sake of speed.  Specifically,
        steps 1 and 3 are done at the same time.

        Step 1: Collect Coefficient Gaps.
                Scan through coefficient list, and record the lengths of
                sequences of zero coefficients.  This corresponds to
                collapsing Horner's Form into a reduced representation.
                For example, ::

                  x^8 + x^4 + x^2 + 1
                   = ((((((((1)*x + 0)*x+0)*x+0)*x+1)*x+0)*x+0)*x+1)*x+0)*x+1
                   = ((((1)*x^4 + 1)*x^2 + 1)*x^2 + 1

                gives a list of "gaps": [2,4]

        Step 2: Fill in Gap Structure.
                Given the list of gaps, find a reasonable sequence of
                of multiplications / squarings of x that will result in
                the computation of all desired exponents.  Record this
                sequence of steps as an evaluation DAG, and retain
                references to the nodes representing the desired
                exponents.  For the example above, we would have::

                  x^2 = x*x
                  x^4 = (x^2) * (x^2)

        Step 3: Construct Evaluation Dag.
                Rescan the coefficient list, and build an evaluation DAG
                representation for the reduced Horner Form as above.
                Whenever a sequence of zeros is encountered, multiply by
                the appropriate "gap" node.  Retain a reference to the
                result node.

        Implementation considerations:

        * By combining steps 1 and 3, we greatly improve the speed of
          this construction, but some complexity is introduced.  The
          solution to this is that "dummy" nodes are created to represent
          the desired gaps.  As the structure of the gaps is filled in,
          these dummies get references to usable DAG nodes.  After all gaps
          are filled in, we strip out dummy nodes, and are left with a
          complete representation of our polynomial.

        * The "binary" algorithm (currently the only algorithm; others are
          forthcoming) requires the gaps to considered in order, and adds
          additional dummies as it goes.  Hence, the gaps are put into a
          binary tree.
        """
        cdef generic_pd max_gap, dag
        cdef BinaryTree gaps

        self._coeffs = coeffs
        gaps, dag = self._parse_structure()
        max_gap = <generic_pd>(gaps.get_max())
        if max_gap.label > 1:
            if algorithm == 'binary':
                self._fill_gaps_binary(gaps)
            elif algorithm == 'pippenger':
                raise NotImplementedError("Implementation of Pippenger's Algorithm is not ready for prime time.")
            else:
                raise RuntimeError("Method '%s' not supported.")

        self._dag = dag.nodummies()

    def __repr__(self):
        return "CompiledPolynomialFunction(%s)" % self._dag

    def __call__(self, x):
        return self.eval(x)

    cdef object eval(CompiledPolynomialFunction self, object x):
        cdef object temp
        try:
            pd_eval(self._dag, x, self._coeffs)  #see further down
            temp = self._dag.value               #for an explanation
            pd_clean(self._dag)                  #of these 3 lines
            return temp
        except TypeError as msg:
            self._dag.reset()
            raise TypeError(msg)

    cdef object _parse_structure(CompiledPolynomialFunction self):
        """
        Loop through the coefficients of the polynomial, and collect
        coefficient gap widths.  Meanwhile, construct the evaluation
        DAG; inserting dummy nodes wherever there are gaps of width
        greater than 1.

        Return the resultant head DAG node, and a binary tree
        containing the dummy nodes.
        """
        cdef BinaryTree gaps
        cdef int d
        cdef generic_pd s

        s = univar_pd()
        gaps = BinaryTree()
        gaps.insert(1, s)

        d = len(self._coeffs)-1

        s = coeff_pd(d)
        gap_width = 0

        d -= 1
        while d > 0:
            gap_width += 1
            if self._coeffs[d]:
                s = abc_pd(s, self._get_gap(gaps, gap_width), d)
                gap_width = 0
            d-=1

        gap_width += 1
        if self._coeffs[0]:
            s = abc_pd(s, self._get_gap(gaps, gap_width), 0)
        else:
            s = mul_pd(s, self._get_gap(gaps, gap_width))

        return gaps, s

    cdef generic_pd _get_gap(CompiledPolynomialFunction self, BinaryTree gaps, int gap):
        """
        Find an entry in the BinaryTree gaps, identified by the int gap.
        If such an entry does not exist, create it and put it in the tree.
        """

        cdef generic_pd g
        g = gaps.get(gap)
        if g is not None:
            return g
        else:
            g = dummy_pd(gap)
            gaps.insert(gap,g)
            return g

    cdef void _fill_gaps_binary(CompiledPolynomialFunction self, BinaryTree gaps) noexcept:
        """
        The desired gaps come in a tree, filled with dummy nodes (with the
        exception of the var node, which is not a dummy).  The nodes are
        labeled with their width.  When we refer to multiplication or
        squaring nodes, we give the current dummy node a usable dag of the
        appropriate type, which treats the nodes being operated on as
        argument.  That is, if we want to multiply nodes A and B, and
        give the result to M, we would call ::

            M.fill(mul_pd(A,B))

        and refer to this operation as ::

            M = A*B

        Sometimes we want a node that isn't in the tree.  In that case, we
        create a new dummy node, and put it into the tree.  The phrase
        "get the node for..." means that we obtain one with the _get_gap
        method, which may create a new node.


        Fill in the gaps with the following algorithm:

        | Step 1:
        |     Remove max node from the tree, denote its width m.
        |     If m == 1, halt.
        |
        | Step 2:
        |     If m is even, check if m/2 is already in our tree.
        |     If yes, square the node corresponding to m/2, and go to step 1.
        |
        | Step 3:
        |     Peek at the next-largest, and denote its width n.
        |     Write m = qn+r.
        |     If r != 0, get the node R, whose gap-width is r.
        |         Also, get the node QN, whose width is qn.
        |         Then, set M = R*QN and go to step 1.
        |     If r == 0, we have two cases:
        |         q is even:
        |             get/create the node for q/n, denote it T.
        |             Set M = T**2 and go to step 1.
        |         q is odd:
        |             get/create the node for (Q-1)*N, denote it T.
        |             Set M = T*N and go to step 1.

        The r == 0 case in step 3 is equivalent to binary exponentiation.
        """
        cdef int m,n,k,r,half
        cdef generic_pd T,N,H
        cdef dummy_pd M
        T = gaps.pop_max()
        while T is not None:
            N = gaps.get_max()
            if N is None:
                return
            else:
                M = <dummy_pd>T
            m = M.label
            n = N.label
            k = m/n
            r = m%n
            half = m/2

            found = False
            if m % 2 == 0 and n >= half:
                H = gaps.get(half)
                if H is not None:
                    M.fill(sqr_pd(H))
                    found = True

            if not found:
                if r > 0:
                    M.fill(mul_pd(self._get_gap(gaps, n*k), self._get_gap(gaps, r)))
                elif k % 2 == 0:
                    M.fill(sqr_pd(self._get_gap(gaps, n*k/2)))
                else:
                    M.fill(mul_pd(self._get_gap(gaps, n*(k-1)), N))

            T = gaps.pop_max()


########################################################
#
#  Polynomial DAG Code
#
# An "evaluation DAG" is a rather generic concept.  What
# follows is a set of DAG classes which are suitable for
# evaluating polynomials in Horner's form, and very
# little else.  Let's call them polydags from here down.
# DAG in all-caps is a whole graph. However,
# polydag or pd := polynomial DAG node
#
# 5 operations, and 3 types of constants are implemented:
#
#  univar_pd: a variable of a univariate polynomial.
#  var_pd: a variable of a multivariate polynomial.
#          parametrized by an index number.
#  const_pd: a coefficient of a polynomial.  Also
#            parametrized by an index number.
#
# unary operations:
#
#  sqr_pd: takes another polydag as an argument, squares
#          by multiplying argument's value by itself.
#  pow_pd: takes another polydag, and an int, as arguments.
#          exponentiates with the ** operator.
#
# binary operations:
#
#  add_pd: adds values of two other polydags.
#  mul_pd: like add_pd, but we with multiplication
#  abc_pd: not technically binary, but it only depends
#          upon two other polydags, so it fits into the
#          class structure this way.  Multiplies the
#          values of two other polydags, and adds the
#          coefficient corresponding to the given index.
#
# Polydag structure:
#  properties:
#    value: some python object.  Should be None before
#           and after the DAG is evaluated.
#    refs: number of polydag nodes that reference self
#    hits: number of times, during current computation,
#          that self.value has been accessed
#    label: used to identify nodes from the outside.
#           specifically, when they're in a binary tree.
#
#  methods:
#    eval: runs the computation for which the node is
#          intended, and stores the result in self.value.
#          Takes 2 python objects as arguments.  The first
#          is the value of the variable, for univariate
#          polynomials, or a tuple of values for multi-
#          variate.  The second is the list of coefficients.
#          The reason for taking the list of coefficients,
#          rather than holding references to them, is that
#          we can implement recursive calling at some later
#          date.
#    nodummies: recursively evict dummies, replacing them
#               with the non-dummy nodes that they
#               reference.


# These inline functions are called wherever a node gets
# evaluated.  First, pd_eval is called to ensure that the
# target DAG node will have its .value property set. It
# serves two purposes:
#   1) Count the number of times this node has been asked
#      for its value.
#   2) Call the node's eval() function if and only if it
#      has not been called yet.  These functions are
#      inlined to prevent excessive function calls.
# Then, pd_clean is called.  It checks to see if all of
# the polydag's dependents have phoned in, and resets
# self.value to None if they have.  Thus, we don't hold
# on to intermediate values any longer than we have to.

cdef inline int pd_eval(generic_pd pd, object vars, object coeffs) except -2:
    if pd.value is None:
        pd.eval(vars, coeffs)
    pd.hits += 1

cdef inline void pd_clean(generic_pd pd) noexcept:
    if pd.hits >= pd.refs:
        pd.value = None
        pd.hits = 0

cdef class generic_pd:
    def __init__(generic_pd self):
        self.value = None
        self.hits = 0
        self.refs = 0
        self.label = -1

    cdef int eval(generic_pd self, object vars, object coeffs) except -2:
        raise NotImplementedError

    cdef generic_pd nodummies(generic_pd self):
        return self

    cdef void reset(generic_pd self) noexcept:
        self.hits = 0
        self.value = None

cdef class dummy_pd(generic_pd):
    def __init__(dummy_pd self, int label):
        self.label = label

    cdef void fill(dummy_pd self, generic_pd link) noexcept:
        self.link = link

    cdef generic_pd nodummies(dummy_pd self):
        #sorry guys, this is my stop
        self.link.refs = self.refs
        return self.link.nodummies()

cdef class var_pd(generic_pd):
    def __init__(var_pd self, int index):
        generic_pd.__init__(self)
        self.index = index
    cdef int eval(var_pd self, object vars, object coeffs) except -2:
        self.value = vars[self.index]

    def __repr__(self):
        return "x[%s]" % (self.index)


cdef class univar_pd(generic_pd):
    def __init__(univar_pd self):
        generic_pd.__init__(self)
        self.label = 1
    cdef int eval(univar_pd self, object var, object coeffs) except -2:
        self.value = var

    def __repr__(self):
        return "x"

cdef class coeff_pd(generic_pd):
    def __init__(coeff_pd self, int index):
        generic_pd.__init__(self)
        self.index = index
    cdef int eval(coeff_pd self, object vars, object coeffs) except -2:
        self.value = coeffs[self.index]

    def __repr__(self):
        return "a%s" % (self.index)

    cdef void reset(self) noexcept:
        pass

cdef class unary_pd(generic_pd):
    def __init__(unary_pd self, generic_pd operand):
        generic_pd.__init__(self)
        self.operand = operand
        self.operand.refs += 1

    cdef generic_pd nodummies(self):
        self.operand = self.operand.nodummies()
        return self

    cdef void reset(self) noexcept:
        generic_pd.reset(self)
        self.operand.reset()


cdef class sqr_pd(unary_pd):
    cdef int eval(sqr_pd self, object vars, object coeffs) except -2:
        pd_eval(self.operand, vars, coeffs)
        self.value = self.operand.value * self.operand.value
        pd_clean(self.operand)

    def __repr__(self):
        return "(%s)^2" % (self.operand)

cdef class pow_pd(unary_pd):
    def __init__(unary_pd self, generic_pd base, object exponent):
        unary_pd.__init__(self, base)
        self.exponent = exponent

    cdef int eval(pow_pd self, object vars, object coeffs) except -2:
        pd_eval(self.operand, vars, coeffs)
        self.value = self.operand.value ** self.exponent
        pd_clean(self.operand)

    def __repr__(self):
        return "(%s^%s)" % (self.left, self.exponent)


cdef class binary_pd(generic_pd):
    def __init__(binary_pd self, generic_pd left, generic_pd right):
        generic_pd.__init__(self)
        self.left = left
        self.right = right
        self.left.refs+= 1
        self.right.refs+= 1

    cdef generic_pd nodummies(self):
        self.left = self.left.nodummies()
        self.right = self.right.nodummies()
        return self

    cdef void reset(self) noexcept:
        generic_pd.reset(self)
        self.left.reset()
        self.right.reset()

cdef class add_pd(binary_pd):
    cdef int eval(add_pd self, object vars, object coeffs) except -2:
        pd_eval(self.left, vars, coeffs)
        pd_eval(self.right, vars, coeffs)
        self.value = self.left.value + self.right.value
        pd_clean(self.left)
        pd_clean(self.right)

    def __repr__(self):
        return "(%s+%s)" % (self.left, self.right)

cdef class mul_pd(binary_pd):
    cdef int eval(mul_pd self, object vars, object coeffs) except -2:
        pd_eval(self.left, vars, coeffs)
        pd_eval(self.right, vars, coeffs)
        self.value = self.left.value * self.right.value
        pd_clean(self.left)
        pd_clean(self.right)

    def __repr__(self):
        return "(%s*%s)" % (self.left, self.right)

cdef class abc_pd(binary_pd):
    def __init__(abc_pd self, generic_pd left, generic_pd right, int index):
        binary_pd.__init__(self, left, right)
        self.index = index

    def __repr__(self):
        return "(%s*%s+a%s)" % (self.left, self.right, self.index)

    cdef int eval(abc_pd self, object vars, object coeffs) except -2:
        pd_eval(self.left, vars, coeffs)
        pd_eval(self.right, vars, coeffs)
        self.value = self.left.value * self.right.value + coeffs[self.index]
        pd_clean(self.left)
        pd_clean(self.right)



# sage_setup: distribution = sagemath-categories
"""
Scheme obtained by gluing two other schemes
"""

#*******************************************************************************
#  Copyright (C) 2006 William Stein
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
#*******************************************************************************

from sage.misc.lazy_import import lazy_import
from sage.schemes.generic.scheme import Scheme

lazy_import('sage.schemes.generic.morphism', 'SchemeMorphism')


class GluedScheme(Scheme):
    r"""
    INPUT:

    - ``f`` -- open immersion from a scheme `U` to a scheme `X`

    - ``g`` -- open immersion from `U` to a scheme `Y`

    OUTPUT: the scheme obtained by gluing `X` and `Y` along the open set `U`

    .. NOTE::

       Checking that `f` and `g` are open
       immersions is not implemented.

    EXAMPLES::

        sage: # needs sage.libs.singular
        sage: R.<x, y> = QQ[]
        sage: S.<xbar, ybar> = R.quotient(x * y - 1)
        sage: Rx = QQ["x"]
        sage: Ry = QQ["y"]
        sage: phi_x = Rx.hom([xbar])
        sage: phi_y = Ry.hom([ybar])
        sage: Sx = Schemes()(phi_x)
        sage: Sy = Schemes()(phi_y)
        sage: Sx.glue_along_domains(Sy)
        Scheme obtained by gluing X and Y along U, where
          X: Spectrum of Univariate Polynomial Ring in x over Rational Field
          Y: Spectrum of Univariate Polynomial Ring in y over Rational Field
          U: Spectrum of Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x*y - 1)
    """
    def __init__(self, f, g, check=True):
        if check:
            if not isinstance(f, SchemeMorphism):
                raise TypeError("f (=%s) must be a scheme morphism" % f)
            if not isinstance(g, SchemeMorphism):
                raise TypeError("g (=%s) must be a scheme morphism" % g)
            if f.domain() != g.domain():
                raise ValueError("f (=%s) and g (=%s) must have the same domain" % (f,g))
        self.__f = f
        self.__g = g

    def gluing_maps(self):
        r"""
        Return the gluing maps of this glued scheme, i.e. the maps `f` and `g`.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: R.<x, y> = QQ[]
            sage: S.<xbar, ybar> = R.quotient(x * y - 1)
            sage: Rx = QQ["x"]
            sage: Ry = QQ["y"]
            sage: phi_x = Rx.hom([xbar])
            sage: phi_y = Ry.hom([ybar])
            sage: Sx = Schemes()(phi_x)
            sage: Sy = Schemes()(phi_y)
            sage: Sxy = Sx.glue_along_domains(Sy)
            sage: Sxy.gluing_maps() == (Sx, Sy)
            True
        """
        return self.__f, self.__g

    def _repr_(self):
        return "Scheme obtained by gluing X and Y along U, where\n  X: %s\n  Y: %s\n  U: %s" % (
            self.__f.codomain(), self.__g.codomain(), self.__f.domain())



# sage_setup: distribution = sagemath-categories
r"""
Orders of function fields

An order of a function field is a subring that is, as a module over the base
maximal order, finitely generated and of maximal rank `n`, where `n` is the
extension degree of the function field. All orders are subrings of maximal
orders.

A rational function field has two maximal orders: maximal finite order `o` and
maximal infinite order `o_\infty`. The maximal order of a rational function
field over constant field `k` is just the polynomial ring `o=k[x]`. The
maximal infinite order is the set of rational functions whose denominator has
degree greater than or equal to that of the numerator.

EXAMPLES::

    sage: K.<x> = FunctionField(QQ)
    sage: O = K.maximal_order()
    sage: I = O.ideal(1/x); I
    Ideal (1/x) of Maximal order of Rational function field in x over Rational Field
    sage: 1/x in O
    False
    sage: Oinf = K.maximal_order_infinite()
    sage: 1/x in Oinf
    True

In an extension of a rational function field, an order over the maximal finite
order is called a finite order while an order over the maximal infinite order
is called an infinite order. Thus a function field has one maximal finite order
`O` and one maximal infinite order `O_\infty`. There are other non-maximal
orders such as equation orders::

    sage: # needs sage.rings.function_field
    sage: K.<x> = FunctionField(GF(3)); R.<y> = K[]
    sage: L.<y> = K.extension(y^3 - y - x)
    sage: O = L.equation_order()
    sage: 1/y in O
    False
    sage: x/y in O
    True

Sage provides an extensive functionality for computations in maximal orders of
function fields. For example, you can decompose a prime ideal of a rational
function field in an extension::

    sage: K.<x> = FunctionField(GF(2)); _.<t> = K[]
    sage: o = K.maximal_order()
    sage: p = o.ideal(x + 1)
    sage: p.is_prime()                                                                  # needs sage.libs.pari
    True

    sage: # needs sage.rings.function_field
    sage: F.<y> = K.extension(t^3 - x^2*(x^2 + x + 1)^2)
    sage: O = F.maximal_order()
    sage: O.decomposition(p)
    [(Ideal (x + 1, y + 1) of Maximal order
     of Function field in y defined by y^3 + x^6 + x^4 + x^2, 1, 1),
     (Ideal (x + 1, (1/(x^3 + x^2 + x))*y^2 + y + 1) of Maximal order
     of Function field in y defined by y^3 + x^6 + x^4 + x^2, 2, 1)]

    sage: # needs sage.rings.function_field
    sage: p1, relative_degree,ramification_index = O.decomposition(p)[1]
    sage: p1.parent()
    Monoid of ideals of Maximal order of Function field in y
    defined by y^3 + x^6 + x^4 + x^2
    sage: relative_degree
    2
    sage: ramification_index
    1

When the base constant field is the algebraic field `\QQbar`, the only prime ideals
of the maximal order of the rational function field are linear polynomials. ::

    sage: # needs sage.rings.function_field sage.rings.number_field
    sage: K.<x> = FunctionField(QQbar)
    sage: R.<y> = K[]
    sage: L.<y> = K.extension(y^2 - (x^3-x^2))
    sage: p = K.maximal_order().ideal(x)
    sage: L.maximal_order().decomposition(p)
    [(Ideal (1/x*y - I) of Maximal order of Function field in y defined by y^2 - x^3 + x^2,
      1,
      1),
     (Ideal (1/x*y + I) of Maximal order of Function field in y defined by y^2 - x^3 + x^2,
      1,
      1)]


AUTHORS:

- William Stein (2010): initial version

- Maarten Derickx (2011-09-14): fixed ideal_with_gens_over_base() for rational function fields

- Julian Rueth (2011-09-14): added check in _element_constructor_

- Kwankyu Lee (2017-04-30): added maximal orders of global function fields

- Brent Baccala (2019-12-20): support orders in characteristic zero
"""

# ****************************************************************************
#       Copyright (C) 2010      William Stein <wstein@gmail.com>
#                     2011      Maarten Derickx <m.derickx.student@gmail.com>
#                     2011      Julian Rueth <julian.rueth@gmail.com>
#                     2017-2020 Kwankyu Lee
#                     2019      Brent Baccala
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.integral_domains import IntegralDomains
from sage.structure.parent import Parent
from sage.structure.unique_representation import CachedRepresentation, UniqueRepresentation

from .ideal import IdealMonoid, FunctionFieldIdeal


class FunctionFieldOrder_base(CachedRepresentation, Parent):
    """
    Base class for orders in function fields.

    INPUT:

    - ``field`` -- function field

    EXAMPLES::

        sage: F = FunctionField(QQ,'y')
        sage: F.maximal_order()
        Maximal order of Rational function field in y over Rational Field
    """
    def __init__(self, field, ideal_class=FunctionFieldIdeal, category=None):
        """
        Initialize.

        TESTS::

            sage: F = FunctionField(QQ,'y')
            sage: O = F.maximal_order()
            sage: TestSuite(O).run(skip='_test_gcd_vs_xgcd')
        """
        category = IntegralDomains().or_subcategory(category).Infinite()
        Parent.__init__(self, category=category, facade=field)

        self._ideal_class = ideal_class # element class for parent ideal monoid
        self._field = field

    def is_field(self, proof=True):
        """
        Return ``False`` since orders are never fields.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order().is_field()
            False
        """
        return False

    def is_noetherian(self):
        """
        Return ``True`` since orders in function fields are Noetherian.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order().is_noetherian()
            True
        """
        return True

    def function_field(self):
        """
        Return the function field to which the order belongs.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order().function_field()
            Rational function field in y over Rational Field
        """
        return self._field

    fraction_field = function_field

    def is_subring(self, other):
        """
        Return ``True`` if the order is a subring of the other order.

        INPUT:

        - ``other`` -- order of the function field or the field itself

        EXAMPLES::

            sage: F = FunctionField(QQ,'y')
            sage: O = F.maximal_order()
            sage: O.is_subring(F)
            True
        """
        if other is self._field:
            return True
        else:
            raise NotImplementedError

    def ideal_monoid(self):
        """
        Return the monoid of ideals of the order.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order().ideal_monoid()
            Monoid of ideals of Maximal order of Rational function field in y over Rational Field
        """
        return IdealMonoid(self)


class FunctionFieldOrder(FunctionFieldOrder_base):
    """
    Base class for orders in function fields.
    """
    def _repr_(self):
        """
        Return the string representation.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order()
            Maximal order of Rational function field in y over Rational Field
        """
        return "Order in {}".format(self._field)


class FunctionFieldOrderInfinite(FunctionFieldOrder_base):
    """
    Base class for infinite orders in function fields.
    """
    def _repr_(self):
        """
        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order_infinite()
            Maximal infinite order of Rational function field in y over Rational Field
        """
        return "Infinite order in {}".format(self.function_field())


class FunctionFieldMaximalOrder(UniqueRepresentation, FunctionFieldOrder):
    """
    Base class of maximal orders of function fields.
    """
    def _repr_(self):
        """
        Return the string representation of the order.

        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order()._repr_()
            'Maximal order of Rational function field in y over Rational Field'
        """
        return "Maximal order of %s" % (self.function_field(),)


class FunctionFieldMaximalOrderInfinite(FunctionFieldMaximalOrder, FunctionFieldOrderInfinite):
    """
    Base class of maximal infinite orders of function fields.
    """
    def _repr_(self):
        """
        EXAMPLES::

            sage: FunctionField(QQ,'y').maximal_order_infinite()
            Maximal infinite order of Rational function field in y over Rational Field

            sage: K.<x> = FunctionField(GF(2)); R.<t> = PolynomialRing(K)
            sage: F.<y> = K.extension(t^3 - x^2*(x^2+x+1)^2)                            # needs sage.rings.function_field
            sage: F.maximal_order_infinite()                                            # needs sage.rings.function_field
            Maximal infinite order of Function field in y defined by y^3 + x^6 + x^4 + x^2
        """
        return "Maximal infinite order of %s" % (self.function_field(),)



# sage_setup: distribution = sagemath-categories
"""
Polynomial Interfaces to Singular

AUTHORS:

- Martin Albrecht <malb@informatik.uni-bremen.de> (2006-04-21)
- Robert Bradshaw: Re-factor to avoid multiple inheritance vs. Cython (2007-09)
- Syed Ahmad Lavasani: Added function field to _singular_init_ (2011-12-16);
  Added non-prime finite fields to _singular_init_ (2012-1-22)

TESTS::

    sage: R = PolynomialRing(GF(2**8,'a'), 10, 'x', order='invlex')                     # needs sage.rings.finite_rings
    sage: R == loads(dumps(R))                                                          # needs sage.rings.finite_rings
    True
    sage: P.<a,b> = PolynomialRing(GF(7), 2)
    sage: f = (a^3 + 2*b^2*a)^7; f
    a^21 + 2*a^7*b^14
"""
#################################################################
#
#   Sage: Open Source Mathematical Software
#
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#
######################################################################

import sage.rings.fraction_field
import sage.rings.abc
import sage.rings.number_field as number_field

from sage.rings.rational_field import RationalField
from sage.rings.function_field.function_field_rational import RationalFunctionField
from sage.rings.finite_rings.finite_field_base import FiniteField
from sage.rings.integer_ring import ZZ
from sage.rings.number_field.number_field_base import NumberField

import sage.rings.finite_rings.finite_field_constructor


def _do_singular_init_(singular, base_ring, char, _vars, order):
    r"""
    Implementation of :meth:`PolynomialRing_singular_repr._singular_init_`.

    This code was extracted from :class:`PolynomialRing_singular_repr`
    to make it callable from other places, in particular
    :class:`MPolynomialRing_libsingular`.

    TESTS::

        sage: from sage.rings.polynomial.polynomial_singular_interface import _do_singular_init_
        sage: _do_singular_init_(singular, ZZ, 0, 'X', 'dp')                            # needs sage.libs.singular
        (polynomial ring, over a domain, global ordering
         // coefficients: ZZ
         // number of vars : 1
         //        block   1 : ordering dp
         //                  : names    X
         //        block   2 : ordering C,
         None)
    """
    if singular is None:
        from sage.interfaces.singular import singular

    make_ring = lambda s: singular.ring(s, _vars, order=order)

    if base_ring is ZZ:
        return make_ring("(ZZ)"), None

    if isinstance(base_ring, sage.rings.rational_field.RationalField):
        return make_ring("(QQ)"), None

    elif isinstance(base_ring, sage.rings.abc.RealField):
        # singular converts to bits from base_10 in mpr_complex.cc by:
        #  size_t bits = 1 + (size_t) ((float)digits * 3.5);
        precision = base_ring.precision()
        digits = (2*precision + 4) // 7
        return make_ring(f"(real,{digits},0)"), None

    elif isinstance(base_ring, sage.rings.abc.ComplexField):
        # singular converts to bits from base_10 in mpr_complex.cc by:
        #  size_t bits = 1 + (size_t) ((float)digits * 3.5);
        precision = base_ring.precision()
        digits = (2*precision + 4) // 7
        return make_ring(f"(complex,{digits},0,I)"), None

    elif isinstance(base_ring, sage.rings.abc.RealDoubleField):
        # singular converts to bits from base_10 in mpr_complex.cc by:
        #  size_t bits = 1 + (size_t) ((float)digits * 3.5);
        return make_ring("(real,15,0)"), None

    elif isinstance(base_ring, sage.rings.abc.ComplexDoubleField):
        # singular converts to bits from base_10 in mpr_complex.cc by:
        #  size_t bits = 1 + (size_t) ((float)digits * 3.5);
        return make_ring("(complex,15,0,I)"), None

    elif isinstance(base_ring, sage.rings.abc.IntegerModRing):
        char = base_ring.characteristic()
        if isinstance(base_ring, FiniteField) and char <= 2147483647:
            return make_ring(str(char)), None
        if char.is_power_of(2):
            return make_ring(f"(integer,2,{char.nbits()-1})"), None
        return make_ring(f"(integer,{char})"), None

    elif isinstance(base_ring, FiniteField):
        # not the prime field!
        gen = str(base_ring.gen())
        R = make_ring(f"({char},{gen})")

        minpoly = str(base_ring.modulus()).replace("x",gen).replace(" ","")
        if singular.eval('minpoly') != f"({minpoly})":
            singular.eval(f"minpoly={minpoly}")
            minpoly = singular.eval('minpoly')[1:-1]

        return R, minpoly

    elif isinstance(base_ring, NumberField) and base_ring.is_absolute():
        # not the rationals!
        gen = str(base_ring.gen())
        poly = base_ring.polynomial()
        poly_gen = str(poly.parent().gen())
        poly_str = str(poly).replace(poly_gen,gen)
        R = make_ring(f"({char},{gen})")

        minpoly = poly_str.replace(" ","")
        if singular.eval('minpoly') != f"({minpoly})":
            singular.eval(f"minpoly={minpoly}")
            minpoly = singular.eval('minpoly')[1:-1]

        return R, minpoly

    elif isinstance(base_ring, sage.rings.fraction_field.FractionField_generic):
        if base_ring.ngens() == 1:
            gens = str(base_ring.gen())
        else:
            gens = str(base_ring.gens())

        B = base_ring.base_ring()
        base_char = base_ring.characteristic()

        if B.is_prime_field() or B is ZZ:
            return make_ring(f"({base_char},{gens})"), None

        if isinstance(B, FiniteField) and B.characteristic() <= 2147483647:
            ext_gen = str(B.gen())
            _vars = '(' + ext_gen + ', ' + _vars[1:]

            R = make_ring(f"({base_char},{gens})")

            base_ring.__minpoly = (str(B.modulus()).replace("x",ext_gen)).replace(" ","")
            singular.eval('setring ' + R._name)

            from sage.misc.stopgap import stopgap
            stopgap("Denominators of fraction field elements are sometimes dropped without warning.", 17696)

            return singular(f"std(ideal({base_ring.__minpoly}))", type='qring'), None

    elif isinstance(base_ring, RationalFunctionField) \
            and base_ring.constant_field().is_prime_field():
        gen = str(base_ring.gen())
        return make_ring(f"({base_ring.characteristic()},{gen})"), None

    raise TypeError("no conversion to a Singular ring defined")


class PolynomialRing_singular_repr:
    """
    Implement methods to convert polynomial rings to Singular.

    This class is a base class for all univariate and multivariate
    polynomial rings which support conversion from and to Singular
    rings.
    """
    def _singular_(self, singular=None):
        r"""
        Return a Singular ring for this polynomial ring.

        Currently `\QQ`, `{\rm GF}(p), {\rm GF}(p^n)`, `\CC`, `\RR`, `\ZZ` and
        `\ZZ/n\ZZ` are supported.

        INPUT:

        - ``singular`` -- Singular instance

        OUTPUT: Singular ring matching this ring

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(CC)                                          # needs sage.rings.real_mpfr
            sage: singular(R)                                                           # needs sage.libs.singular sage.rings.real_mpfr
            polynomial ring, over a field, global ordering
            // coefficients: real[I](complex:15 digits, additional 0 digits)/(I^2+1)
            // number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C

            sage: R.<x,y> = PolynomialRing(RealField(100))                              # needs sage.rings.real_mpfr
            sage: singular(R)                                                           # needs sage.libs.singular sage.rings.real_mpfr
            polynomial ring, over a field, global ordering
            // coefficients: Float()
            // number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C

            sage: w = polygen(ZZ, 'w')
            sage: R.<x> = PolynomialRing(NumberField(w^2 + 1, 's'))                     # needs sage.rings.number_field
            sage: singular(R)                                                           # needs sage.libs.singular sage.rings.number_field
            polynomial ring, over a field, global ordering
            //   coefficients: QQ[s]/(s^2+1)
            //   number of vars : 1
            //        block   1 : ordering lp
            //                  : names    x
            //        block   2 : ordering C

            sage: R = PolynomialRing(GF(127), 'x', implementation='singular')           # needs sage.libs.singular
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: ZZ/127
            //   number of vars : 1
            //        block   1 : ordering dp
            //                  : names    x
            //        block   2 : ordering C

            sage: R = PolynomialRing(QQ, 'x', implementation='singular')                # needs sage.libs.singular
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: QQ
            //   number of vars : 1
            //        block   1 : ordering dp
            //                  : names    x
            //        block   2 : ordering C

            sage: R = PolynomialRing(QQ,'x')
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: QQ
            //   number of vars : 1
            //        block   1 : ordering lp
            //                  : names    x
            //        block   2 : ordering C

            sage: R = PolynomialRing(GF(127), 'x')
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: ZZ/127
            //   number of vars : 1
            //        block   1 : ordering lp
            //                  : names    x
            //        block   2 : ordering C

            sage: R = Frac(ZZ['a,b'])['x,y']                                            # needs sage.libs.singular
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: QQ(a, b)
            //   number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C


            sage: R = IntegerModRing(1024)['x,y']
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a ring (with zero-divisors), global ordering
            //   coefficients: ZZ/(2^10)
            //   number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C

            sage: R = IntegerModRing(15)['x,y']
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a ring (with zero-divisors), global ordering
            //   coefficients: ZZ/...(15)
            //   number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C

            sage: R = ZZ['x,y']
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a domain, global ordering
            //   coefficients: ZZ
            //   number of vars : 2
            //        block   1 : ordering dp
            //                  : names    x y
            //        block   2 : ordering C

            sage: R = ZZ['x']
            sage: singular(R)                                                           # needs sage.libs.singular
            polynomial ring, over a domain, global ordering
            // coefficients: ZZ
            // number of vars : 1
            //        block   1 : ordering lp
            //                  : names    x
            //        block   2 : ordering C

            sage: # needs sage.rings.finite_rings
            sage: k.<a> = FiniteField(25)
            sage: R = k['x']
            sage: K = R.fraction_field()
            sage: S = K['y']
            sage: singular(S)                                                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: ZZ/5(x)
            //   number of vars : 2
            //        block   1 : ordering lp
            //                  : names    a y
            //        block   2 : ordering C
            // quotient ring from ideal
            _[1]=a2-a+2

        .. warning::

            - If the base ring is a finite extension field or a number field
              the ring will not only be returned but also be set as the current
              ring in Singular.
            - Singular represents precision of floating point numbers base 10
              while Sage represents floating point precision base 2.
        """
        if singular is None:
            from sage.interfaces.singular import singular
        try:
            R = self.__singular
            if singular is not None and R.parent() is not singular:
                raise ValueError
            elif singular is None:
                from sage.interfaces.singular import singular
            R._check_valid()
            if self.base_ring() is ZZ or self.base_ring().is_prime_field():
                return R
            if isinstance(self.base_ring(), FiniteField) or \
                    (isinstance(self.base_ring(), NumberField) and self.base_ring().is_absolute()):
                R.set_ring()  # sorry for that, but needed for minpoly
                if singular.eval('minpoly') != f"({self.__minpoly})":
                    singular.eval(f"minpoly={self.__minpoly}")
                    self.__minpoly = singular.eval('minpoly')[1:-1]
            return R
        except (AttributeError, ValueError):
            return self._singular_init_(singular)

    def _singular_init_(self, singular=None):
        """
        Return a newly created Singular ring matching this ring.

        EXAMPLES::

            sage: PolynomialRing(QQ,'u_ba')._singular_init_()                           # needs sage.libs.singular
            polynomial ring, over a field, global ordering
            //   coefficients: QQ
            //   number of vars : 1
            //        block   1 : ordering lp
            //                  : names    u_ba
            //        block   2 : ordering C
        """
        if not can_convert_to_singular(self):
            raise TypeError("no conversion of this ring to a Singular ring defined")

        if self.ngens() == 1:
            _vars = f'({self.gen()})'
            if "*" in _vars:  # 1.000...000*x
                _vars = _vars.split("*")[1]
            order = 'lp'
        else:
            _vars = str(self.gens())
            order = self.term_order().singular_str()

        if singular is None:
            from sage.interfaces.singular import singular

        self.__singular, self.__minpoly = _do_singular_init_(singular, self.base_ring(), self.characteristic(), _vars, order)

        return self.__singular


def can_convert_to_singular(R):
    """
    Return ``True`` if this ring's base field or ring can be
    represented in Singular, and the polynomial ring has at
    least one generator.

    The following base rings are supported: finite fields,
    rationals, number fields, and real and complex fields.

    EXAMPLES::

        sage: from sage.rings.polynomial.polynomial_singular_interface import can_convert_to_singular
        sage: can_convert_to_singular(PolynomialRing(QQ, names=['x']))
        True
        sage: can_convert_to_singular(PolynomialRing(ZZ, names=['x']))
        True

        sage: can_convert_to_singular(PolynomialRing(QQ, names=[]))
        False

    TESTS:

    Avoid non absolute number fields (see :issue:`23535`)::

        sage: x = polygen(ZZ, 'x')
        sage: K.<a,b> = NumberField([x^2 - 2, x^2 - 5])                                 # needs sage.rings.number_field
        sage: can_convert_to_singular(K['s,t'])                                         # needs sage.rings.number_field
        False

    Check for :issue:`33319`::

        sage: # needs sage.rings.finite_rings
        sage: R.<x,y> = GF((2^31-1)^3)[]
        sage: R._has_singular
        True
        sage: R.<x,y> = GF((2^31+11)^2)[]
        sage: R._has_singular
        False
        sage: R.<x,y> = GF(10^20 - 11)[]
        sage: R._has_singular
        True

        sage: R.<x,y> = Zmod(10^20 + 1)[]
        sage: R._has_singular
        True
    """
    if R.ngens() == 0:
        return False

    base_ring = R.base_ring()
    if (base_ring is ZZ
        or isinstance(base_ring, RationalField)
        or isinstance(base_ring, (sage.rings.abc.IntegerModRing,
                                  sage.rings.abc.RealField, sage.rings.abc.ComplexField,
                                  sage.rings.abc.RealDoubleField, sage.rings.abc.ComplexDoubleField))):
        return True
    elif isinstance(base_ring, FiniteField):
        return base_ring.characteristic() <= 2147483647
    elif isinstance(base_ring, NumberField):
        return base_ring.is_absolute()
    elif isinstance(base_ring, sage.rings.fraction_field.FractionField_generic):
        B = base_ring.base_ring()
        return (B.is_prime_field() or B is ZZ
                or (isinstance(B, FiniteField) and B.characteristic() <= 2147483647))
    elif isinstance(base_ring, RationalFunctionField):
        return base_ring.constant_field().is_prime_field()
    else:
        return False


class Polynomial_singular_repr:
    """
    Implement coercion of polynomials to Singular polynomials.

    This class is a base class for all (univariate and multivariate)
    polynomial classes which support conversion from and to
    Singular polynomials.

    Due to the incompatibility of Python extension classes and multiple inheritance,
    this just defers to module-level functions.
    """
    def _singular_(self, singular=None):
        if singular is None:
            from sage.interfaces.singular import singular
        return _singular_func(self, singular)

    def _singular_init_func(self, singular=None):
        if singular is None:
            from sage.interfaces.singular import singular
        return _singular_init_func(self, singular)


def _singular_func(self, singular=None):
    """
    Return Singular polynomial matching this polynomial.

    INPUT:

    - ``singular`` -- Singular instance to use

    EXAMPLES::

        sage: # needs sage.libs.singular
        sage: P.<a,b> = PolynomialRing(GF(7), 2)
        sage: f = (a^3 + 2*b^2*a)^7; f
        a^21 + 2*a^7*b^14
        sage: h = f._singular_(); h
        a^21+2*a^7*b^14
        sage: P(h)
        a^21 + 2*a^7*b^14
        sage: P(h^20) == f^20
        True

        sage: # needs sage.libs.singular
        sage: R.<x> = PolynomialRing(GF(7))
        sage: f = (x^3 + 2*x^2*x)^7
        sage: f
        3*x^21
        sage: h = f._singular_(); h
        3*x^21
        sage: R(h)
        3*x^21
        sage: R(h^20) == f^20
        True
    """
    if singular is None:
        from sage.interfaces.singular import singular
    self.parent()._singular_(singular).set_ring()  # this is expensive
    try:
        self.__singular._check_valid()
        if self.__singular.parent() is singular:
            return self.__singular
    except (AttributeError, ValueError):
        pass
    return _singular_init_func(self, singular)


def _singular_init_func(self, singular=None):
    """
    Return corresponding Singular polynomial but enforce that a new
    instance is created in the Singular interpreter.

    Use ``self._singular_()`` instead.
    """
    if singular is None:
        from sage.interfaces.singular import singular
    self.parent()._singular_(singular).set_ring()  # this is expensive
    self.__singular = singular(str(self))
    return self.__singular



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.schemes
r"""
Jacobians in Hess model

This module implements Jacobian arithmetic based on divisor representation by
ideals. This approach to Jacobian arithmetic implementation is attributed to
Hess [Hes2002]_.

Jacobian
--------

To create a Jacobian in Hess model, specify ``'hess'`` model and provide a base divisor
of degree `g`, which is the genus of the function field::

    sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
    sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
    sage: C.geometric_genus()
    1
    sage: B = C([0,1,0]).place()
    sage: B.degree()
    1
    sage: J = C.jacobian(model='hess', base_div=B)
    sage: J
    Jacobian of Projective Plane Curve over Finite Field of size 29
     defined by x^3 - y^2*z + 5*z^3 (Hess model)

Group of rational points
------------------------

The group of rational points of a Jacobian is created from the Jacobian. A
point of the Jacobian group is determined by a divisor (of degree zero) of the
form `D - B` where `D` is an effective divisor of degree `g` and `B` is the base
divisor. Hence a point of the Jacobian group is represented by `D`.

::

    sage: G = J.group()
    sage: P1 = C([1,8,1]).place()
    sage: P2 = C([2,10,1]).place()
    sage: p1 = G(P1)
    sage: p2 = G(P2)
    sage: p1
    [Place (y + 21, z + 28)]
    sage: p2
    [Place (y + 24, z + 14)]
    sage: p1 + p2
    [Place (y + 8, z + 28)]

AUTHORS:

- Kwankyu Lee (2022-01-24): initial version
"""

# ****************************************************************************
#       Copyright (C) 2022 Kwankyu Lee <ekwankyu@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp import op_EQ, richcmp

from sage.categories.map import Map
from sage.categories.commutative_additive_groups import CommutativeAdditiveGroups
from sage.categories.homset import Hom

from sage.arith.misc import integer_ceil
from sage.arith.functions import lcm

from sage.rings.integer import Integer
from sage.matrix.constructor import matrix

from sage.combinat.integer_vector_weighted import WeightedIntegerVectors

from .place import FunctionFieldPlace
from .divisor import FunctionFieldDivisor

from .jacobian_base import (Jacobian_base,
                            JacobianGroup_base,
                            JacobianGroup_finite_field_base,
                            JacobianPoint_base,
                            JacobianPoint_finite_field_base)


class JacobianPoint(JacobianPoint_base):
    """
    Points of Jacobians represented by a pair of ideals.

    If a point of Jacobian is determined by `D`, then the divisor `D` is
    represented by a pair of ideals in the finite maximal order and the
    infinite maximal order of the function field.

    For efficiency reasons, the actual ideals stored are the inverted ideals
    of the ideals representing the divisor `D`.

    INPUT:

    - ``parent`` -- Jacobian group

    - ``dS`` -- an ideal of the finite maximal order of a function field

    - ``ds`` -- an ideal of infinite maximal order of a function field

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: G = C.jacobian(model='hess', base_div=b).group()
        sage: pl = C([1,8,1]).place()
        sage: p = G.point(pl - b)
        sage: dS, ds = p._data
        sage: -(dS.divisor() + ds.divisor()) == pl
        True
    """
    def __init__(self, parent, dS, ds):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl = C([1,8,1]).place()
            sage: p = G.point(pl - b)
            sage: TestSuite(p).run(skip=['_test_category','_test_pickling'])
        """
        super().__init__(parent)
        self._data = (dS, ds)

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: G.zero()
            [Place (1/y, 1/y*z)]
        """
        dS, ds = self._data
        divisor = (~dS).divisor() + (~ds).divisor()
        return f'[{divisor}]'

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: K.<x> = FunctionField(GF(2)); _.<Y> = K[]
            sage: F.<y> = K.extension(Y^3 - x^2*(x^2 + x + 1)^2)
            sage: f = x/(y + 1)
            sage: d = f.divisor()
            sage: {d: 1}
            {Place (1/x, 1/x^4*y^2 + 1/x^2*y + 1)
              + Place (1/x, 1/x^2*y + 1)
              + 3*Place (x, (1/(x^3 + x^2 + x))*y^2)
              - 6*Place (x + 1, y + 1): 1}
        """
        return hash(self._data)

    def _richcmp_(self, other, op):
        """
        Compare ``self`` with ``other`` with respect to operator ``op``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([2,10,1]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1 == p1
            True
            sage: p1 != p2
            True
            sage: p1 > p1
            False
            sage: p1 > p2
            False
            sage: p1 < p2
            True
        """
        if op is op_EQ:
            J = self.parent()
            idS, ids = self._data
            jdS, jds = other._data
            return J._normalize(idS / jdS, ids / jds) is not None
        else:
            return richcmp(self._data, other._data, op)

    def _add_(self, other):
        """
        Return the sum of ``self`` and ``other``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([2,10,1]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1 + p2
            [Place (y + 8, z + 3)]
            sage: p1 + p2 == p2 + p1
            True
        """
        G = self.parent()
        idS, ids = self._data
        jdS, jds = other._data
        bdS, bds = G._base_point
        dS, ds = G._normalize(idS * jdS * bdS, ids * jds * bds)
        return G.element_class(self.parent(), dS, ds)

    def _neg_(self):
        """
        Return the negative of this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl = C([-1,2,1]).place()
            sage: p = G.point(pl - b)
            sage: -p
            [Place (y + 27, z + 1)]
            sage: -(-p) == p
            True
        """
        G = self.parent()
        idS, ids = self._data
        bdS2, bds2 = G._base_point_double
        dS, ds = G._normalize(~(idS * bdS2), ~(ids * bds2))
        return G.element_class(self.parent(), dS, ds)

    def multiple(self, n):
        """
        Return the ``n``-th multiple of this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl = C([-1,2,1]).place()
            sage: p = G.point(pl - b)
            sage: p.multiple(100)
            [Place (1/y, 1/y*z + 8)]
        """
        if n == 0:
            return self.parent().zero()

        G = self.parent()
        idS, ids = self._data
        bdS, bds = G._base_point
        bdS2, bds2 = G._base_point_double
        idSbdS2 = idS * bdS2
        idsbds2 = ids * bds2

        if n < 0:
            bits = Integer(-n).digits(2)
        else:
            bits = Integer(n).digits(2)
        bits.pop()

        dS = idS
        ds = ids
        for i in range(len(bits)):
            b = bits.pop()
            if b > 0:
                dS, ds = G._normalize(dS * dS * idSbdS2 , ds * ds * idsbds2)
            else:
                dS, ds = G._normalize(dS * dS * bdS, ds * ds * bds)
        if n < 0:
            dS, ds = G._normalize(~(dS * bdS2), ~(ds * bds2))

        return G.element_class(self.parent(), dS, ds)

    def addflip(self, other):
        """
        Return the addflip of this and ``other`` point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([2,19,1]).place()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1.addflip(p2)
            [Place (y + 8, z + 27)]
            sage: _ == -(p1 + p2)
            True
        """
        return -(self + other)

    def defining_divisor(self):
        """
        Return the effective divisor that defines this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl = C([-1,2,1]).place()
            sage: p = G.point(pl - b)
            sage: p.defining_divisor() == pl
            True
        """
        dS, ds = self._data
        return (~dS).divisor() + (~ds).divisor()

    def order(self, bound=None):
        """
        Return the order of this point.

        ALGORITHM: Shanks' Baby Step Giant Step

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: p = C([-1,2,1]).place()
            sage: pt = G.point(p - b)
            sage: pt.order()
            30
        """
        if bound is None:  # naive
            J = self.parent()
            zero = J.zero()

            m = self
            r = 1
            while m != zero:
                m = m + self
                r += 1
            return r

        # if bound is given, deploy Shanks' Baby Step Giant Step

        J = self.parent()
        B = J.bound_on_order()
        q = integer_ceil(B.sqrt())
        zero = J.zero()

        # baby steps
        b = [zero]
        g = self
        for i in range(q - 1):
            if g == zero:
                return i + 1
            b.append(g)
            g = g + self

        # giant steps
        g0 = (-q)*(self)
        g = g0
        for i in range(q - 1):
            for r in range(q):
                if g == b[r]:
                    return q * (i + 1) + r
            g = g + g0

        # order is neither smaller or nor larger than this
        return q**2

    def divisor(self):
        """
        Return the divisor representing this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: pl = C([-1,2,1]).place()
            sage: p = G.point(pl - b)
            sage: G.point(p.divisor()) == p
            True
        """
        J = self.parent()
        dS, ds = self._data
        return (~dS).divisor() + (~ds).divisor() - J._base_div


class JacobianPoint_finite_field(JacobianPoint, JacobianPoint_finite_field_base):
    """
    Points of Jacobians over finite fields
    """
    pass


class JacobianGroupEmbedding(Map):
    """
    Embeddings between Jacobian groups.

    INPUT:

    - ``base_group`` -- Jacobian group over a base field

    - ``extension_group`` -- Jacobian group over an extension field

    EXAMPLES::

        sage: k = GF(17)
        sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: J = C.jacobian(model='hess', base_div=b)
        sage: G1 = J.group()
        sage: K = k.extension(3)
        sage: G3 = J.group(K)
        sage: G3.coerce_map_from(G1)
        Jacobian group embedding map:
          From: Group of rational points of Jacobian
           over Finite Field of size 17 (Hess model)
          To:   Group of rational points of Jacobian
           over Finite Field in z3 of size 17^3 (Hess model)
    """
    def __init__(self, base_group, extension_group):
        """
        Initialize.

        TESTS::

            sage: k = GF(17)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G1 = J.group()
            sage: K = k.extension(3)
            sage: G3 = J.group(K)
            sage: map = G3.coerce_map_from(G1)
            sage: TestSuite(map).run(skip=['_test_category', '_test_pickling'])
        """
        F = base_group._function_field
        F_base = F.base_field()
        K = F.constant_base_field()

        F_ext = extension_group._function_field
        F_ext_base = F_ext.base_field()
        K_ext = F_ext.constant_base_field()

        # construct embedding of F into F_ext
        embedK = K_ext.coerce_map_from(K)
        embedF_base = F_base.hom(F_ext_base.gen(), embedK)
        if F.degree() > 1:
            embedF = F.hom(F_ext.gen(), embedF_base)
        else:
            embedF = embedF_base

        self._embedF = embedF
        self._O_ext = F_ext.maximal_order()
        self._Oinf_ext = F_ext.maximal_order_infinite()

        Map.__init__(self, Hom(base_group, extension_group, CommutativeAdditiveGroups()))

    def _repr_type(self):
        """
        Return string representation of ``self``.

        TESTS::

            sage: k = GF(17)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G1 = J.group()
            sage: K = k.extension(3)
            sage: G3 = J.group(K)
            sage: G3.coerce_map_from(G1)  # indirect doctest
            Jacobian group embedding map:
              From: Group of rational points of Jacobian
               over Finite Field of size 17 (Hess model)
              To:   Group of rational points of Jacobian
               over Finite Field in z3 of size 17^3 (Hess model)
        """
        return 'Jacobian group embedding'

    def _call_(self, x):
        """
        Conorm map from F to F_ext.

        TESTS::

            sage: k = GF(7)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^5 - x^3 - 2*x - 1).projective_closure()
            sage: J = C.jacobian(model='hess')
            sage: G1 = J.group()
            sage: K = k.extension(3)
            sage: G3 = J.group(K)
            sage: m = G3.coerce_map_from(G1)
            sage: m(G1.zero()) == G3.zero()
            True
        """
        embedF = self._embedF
        O_ext = self._O_ext
        Oinf_ext = self._Oinf_ext

        idS,ids = x._data
        dS = O_ext.ideal([embedF(g) for g in idS.gens()])
        ds = Oinf_ext.ideal([embedF(g) for g in ids.gens()])
        return self.codomain().element_class(self.codomain(), dS, ds)


class JacobianGroup(UniqueRepresentation, JacobianGroup_base):
    """
    Groups of rational points of a Jacobian.

    INPUT:

    - ``parent`` -- a Jacobian

    - ``function_field`` -- a function field

    - ``base_div`` -- an effective divisor of the function field

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: J = C.jacobian(model='hess', base_div=b)
        sage: J.group()
        Group of rational points of Jacobian
         over Finite Field of size 17 (Hess model)
    """
    Element = JacobianPoint
    _embedding_map_class = JacobianGroupEmbedding

    def __init__(self, parent, function_field, base_div):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: TestSuite(G).run(skip=['_test_elements', '_test_pickling'])
        """
        super().__init__(parent, function_field, base_div)

        bdS, bds = self._get_dS_ds(-base_div)
        try:
            bdS._gens_two()  # speed up multiplication with these ideals
            bds._ideal._gens_two()  # by storing vector forms of two generators
        except AttributeError:
            pass
        self._base_point = (bdS, bds)
        self._base_point_double = (bdS * bdS, bds * bds)

        self._base_place = None

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: J.group()
            Group of rational points of Jacobian
             over Finite Field of size 17 (Hess model)
        """
        r = super()._repr_()
        return r + ' (Hess model)'

    def _element_constructor_(self, x):
        """
        Construct an element of ``self`` from ``x``.

        If ``x`` is an effective divisor, then it is assumed to of
        degree `g`, the genus of the function field.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: G = C.jacobian(model='hess', base_div=b).group()
            sage: G(0)
            [Place (1/y, 1/y*z)]
        """
        if x == 0:
            return self.zero()

        if isinstance(x, FunctionFieldPlace):
            if (self._base_place is not None
                and x in self._function_field.place_set()
                and x.degree() == 1):
                x = x - self._base_place
            else:
                x = x.divisor()

        if (isinstance(x, FunctionFieldDivisor)
            and x in self._function_field.divisor_group()):
            if x.degree() == 0:
                return self.point(x)
            if x.is_effective():
                return self.element_class(self, *self._get_dS_ds(x))

        raise ValueError(f"Cannot construct a point from {x}")

    def _get_dS_ds(self, divisor):
        """
        Return (dS,ds) representation of the divisor.

        TESTS::

            sage: k = GF(17)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: pl = C([2,8,1]).place()
            sage: dS, ds = G._get_dS_ds(2*pl)
            sage: (~dS).divisor() + (~ds).divisor() == 2*pl
            True
        """
        F = self._function_field
        O = F.maximal_order()
        Oinf = F.maximal_order_infinite()

        I = O.ideal(1)
        J = Oinf.ideal(1)
        for p in divisor._data:
            m = divisor._data[p]
            if p.is_infinite_place():
                J *= p.prime_ideal() ** (-m)
            else:
                I *= p.prime_ideal() ** (-m)

        return I, J

    def _normalize(self, I, J):
        """
        Return a pair of normalized ideals from `I` and `J`.

        INPUT:

        - ``I`` -- an ideal of the finite maximal order

        - ``J`` -- an ideal of the infinite maximal order

        The output represents an effective divisor linearly equivalent to the
        divisor represented by the given ideals `I` and `J`.

        ALGORITHM:

        Computes a function `f` in the Riemann-Roch space of the divisor `D`
        represented by the (inverted) ideals `I` and `J`. The output is the
        pair of the (inverted) ideals representing the effective divisor `(f) + D`,
        which is linearly equivalent to `D`.

        TESTS::

            sage: k = GF(17)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: pl = C([2,8,1]).place()
            sage: p = G.point(pl - b)
            sage: dS, ds = (p + p)._data  # indirect doctest
            sage: G.point((~dS).divisor() + (~ds).divisor() - b) == p + p
            True
        """
        F = self._function_field
        n = F.degree()

        O = F.maximal_order()
        Oinf = F.maximal_order_infinite()

        # Step 1: construct matrix M of rational functions in x such that
        # M * B == C where B = [b1,b1,...,bn], C =[v1,v2,...,vn]
        V,fr,to = F.free_module(map=True)
        B = matrix([to(b) for b in J.gens_over_base()])
        C = matrix([to(v) for v in I.gens_over_base()])
        M = C * B.inverse()

        # Step 2: get the denominator d of M and set mat = d * M
        den = lcm([e.denominator() for e in M.list()])
        R = den.parent() # polynomial ring
        one = R.one()
        mat = matrix(R, n, [e.numerator() for e in (den*M).list()])
        gens = list(I.gens_over_base())

        # Step 3: transform mat to a weak Popov form, together with gens

        # initialise pivot_row and conflicts list
        found = None
        pivot_row = [[] for i in range(n)]
        conflicts = []
        for i in range(n):
            bestp = -1
            best = -1
            for c in range(n):
                d = mat[i,c].degree()
                if d >= best:
                    bestp = c
                    best = d

            if best <= den.degree():
                found = i
                break

            if best >= 0:
                pivot_row[bestp].append((i,best))
                if len(pivot_row[bestp]) > 1:
                    conflicts.append(bestp)

        if found is None:
            # while there is a conflict, do a simple transformation
            while conflicts:
                c = conflicts.pop()
                row = pivot_row[c]
                i,ideg = row.pop()
                j,jdeg = row.pop()

                if jdeg > ideg:
                    i,j = j,i
                    ideg,jdeg = jdeg,ideg

                coeff = - mat[i,c].lc() / mat[j,c].lc()
                s = coeff * one.shift(ideg - jdeg)

                mat.add_multiple_of_row(i, j, s)
                gens[i] += s * gens[j]

                row.append((j,jdeg))

                bestp = -1
                best = -1
                for c in range(n):
                    d = mat[i,c].degree()
                    if d >= best:
                        bestp = c
                        best = d

                if best <= den.degree():
                    found = i
                    break

                if best >= 0:
                    pivot_row[bestp].append((i,best))
                    if len(pivot_row[bestp]) > 1:
                        conflicts.append(bestp)
            else:
                return None

        f = gens[found]
        return (O.ideal(~f) * I, Oinf.ideal(~f) * J)

    def point(self, divisor):
        """
        Return the point represented by the divisor of degree zero.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: p = C([-1,2,1]).place()
            sage: G.point(p - b)
            [Place (y + 2, z + 1)]
        """
        c = divisor + self._base_div
        f = c.basis_function_space()[0]
        d = f.divisor() + c
        dS, ds = self._get_dS_ds(d)
        return self.element_class(self, dS, ds)

    @cached_method
    def zero(self):
        """
        Return the zero element of this group.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: G.zero()
            [Place (1/y, 1/y*z)]
        """
        bdS,bds = self._base_point
        return self.element_class(self, ~bdS, ~bds)


class JacobianGroup_finite_field(JacobianGroup, JacobianGroup_finite_field_base):
    """
    Jacobian groups of function fields over finite fields.

    INPUT:

    - ``parent`` -- a Jacobian

    - ``function_field`` -- a function field

    - ``base_div`` -- an effective divisor of the function field

    EXAMPLES::

        sage: k = GF(17)
        sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: J = C.jacobian(model='hess', base_div=b)
        sage: G1 = J.group()
        sage: K = k.extension(3)
        sage: G3 = J.group(K)
        sage: G3.coerce_map_from(G1)
        Jacobian group embedding map:
          From: Group of rational points of Jacobian
           over Finite Field of size 17 (Hess model)
          To:   Group of rational points of Jacobian
           over Finite Field in z3 of size 17^3 (Hess model)
    """
    Element = JacobianPoint_finite_field

    def __init__(self, parent, function_field, base_div):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: G = J.group()
            sage: TestSuite(G).run(skip=['_test_elements','_test_pickling'])
        """
        super().__init__(parent, function_field, base_div)

        F = self._function_field
        K = F.constant_base_field()

        R = F.base_field()  # base rational function field
        x = R.gen()
        y = F.gen()

        r = self._parent._function_field.constant_base_field().degree()
        frob_K = K.frobenius_endomorphism(r)
        frob_R = R.hom(x, base_morphism=frob_K)
        frob_F = F.hom(y, base_morphism=frob_R)

        self._frobenius = frob_F

    def __iter__(self):
        """
        Return generator of points of this group.

        TESTS::

            sage: k = GF(7)
            sage: A.<x,y> = AffineSpace(k,2)
            sage: C = Curve(y^2 + x^3 + 2*x + 1).projective_closure()
            sage: J = C.jacobian(model='hess')
            sage: G = J.group()
            sage: len([pt for pt in G])
            11
        """
        g = self._parent._function_field.genus()
        F = self._function_field
        O = F.maximal_order()
        Oinf = F.maximal_order_infinite()

        deg = 1
        support = []
        degrees = []
        multiples = []
        lst = []

        places_infinite = F.places_infinite()
        generators = [iter(places_infinite)]
        num_of_infinite_places = len(places_infinite)
        while True:
            while True:
                try:
                    new_pl = next(generators[-1])
                    break
                except StopIteration:
                    if deg > g:
                        return
                    generators.append(F._places_finite(deg))
                    deg += 1
            multiples.append((g + 1)*[None])
            P = ~new_pl.prime_ideal()
            dn = new_pl.degree()
            I0 = O.ideal(1)
            J0 = Oinf.ideal(1)
            dr = 0
            for r in range(1, g // new_pl.degree() + 1):
                if new_pl.is_infinite_place():
                    J0 = J0 * P
                else:
                    I0 = I0 * P
                multiples[-1][r] = (I0, J0)
                dr = dr + dn
                for weights in WeightedIntegerVectors(g - dr, degrees):
                    I = I0
                    J = J0
                    for i in range(len(support)):
                        w = weights[i]
                        if w > 0:
                            dS, ds = multiples[i][w]
                            if i < num_of_infinite_places:
                                J *= ds  # dS is the unit ideal
                            else:
                                I *= dS  # ds is the unit ideal
                    pt = self.element_class(self, I, J)
                    if pt not in lst:
                        lst.append(pt)
                        yield pt
            support.append(new_pl)
            degrees.append(new_pl.degree())

    def _frobenius_on(self, pt):
        """
        Return the image of ``pt`` acted by the Frobenius automorphism.

        EXAMPLES::

            sage: k = GF(7)
            sage: A.<x,y> = AffineSpace(k,2)
            sage: C = Curve(y^2 + x^3 + 2*x + 1).projective_closure()
            sage: J = C.jacobian(model='hess')
            sage: G1 = J.group()
            sage: G1.order()
            11
            sage: K = k.extension(3)
            sage: G3 = J.group(K)
            sage: pts1 = G1.get_points(11)
            sage: pts3 = G3.get_points(12)
            sage: pt = next(pt for pt in pts3 if pt not in pts1)
            sage: pt.frobenius().frobenius().frobenius() == pt  # indirect doctest
            True
            sage: pt.frobenius() == pt
            False
        """
        frob_F = self._frobenius

        F = self._function_field
        O = F.maximal_order()
        Oinf = F.maximal_order_infinite()

        idS,ids = pt._data
        dS = O.ideal([frob_F(g) for g in idS.gens()])
        ds = Oinf.ideal([frob_F(g) for g in ids.gens()])
        return self.element_class(self, dS, ds)


class Jacobian(Jacobian_base, UniqueRepresentation):
    """
    Jacobians of function fields.

    EXAMPLES::

        sage: k = GF(17)
        sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: C.jacobian(model='hess', base_div=b)
        Jacobian of Projective Plane Curve over Finite Field of size 17
         defined by x^3 - y^2*z + 5*z^3 (Hess model)
    """
    def __init__(self, function_field, base_div, **kwds):
        """
        Initialize.

        TESTS::

            sage: k = GF(17)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: J = C.jacobian(model='hess', base_div=b)
            sage: TestSuite(J).run(skip=['_test_elements','_test_pickling'])
        """
        super().__init__(function_field, base_div, **kwds)

        if function_field.constant_base_field().is_finite():
            self._group_class = JacobianGroup_finite_field
        else:
            self._group_class = JacobianGroup

    def _repr_(self):
        """
        Return the string representation of ``self``.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: C.jacobian(model='hess', base_div=b)
            Jacobian of Projective Plane Curve over Finite Field of size 17
             defined by x^3 - y^2*z + 5*z^3 (Hess model)
        """
        r = super()._repr_()
        return r + ' (Hess model)'



# sage_setup: distribution = sagemath-categories
"""
CPU Detection
"""

import os


def ncpus():
    """
    Return the number of available CPUs in the system.

    ALGORITHM: :func:`os.sched_getaffinity` or :func:`os.cpu_count`

    EXAMPLES::

        sage: sage.parallel.ncpus.ncpus()  # random output -- depends on machine
        2
    """
    # Support Sage environment variable SAGE_NUM_THREADS
    # NOTE: while doctesting, this is forced to be 2 by the
    # sage-runtests script
    try:
        n = os.environ["SAGE_NUM_THREADS"]
    except KeyError:
        pass
    else:
        return int(n)

    n = None

    if hasattr(os, 'sched_getaffinity'):
        n = len(os.sched_getaffinity(0))

    return n or os.cpu_count() or 1



# sage_setup: distribution = sagemath-categories
from sage.misc.binary_tree import BinaryTree
from sage.misc.binary_tree cimport BinaryTree


cdef class generic_pd

cdef class CompiledPolynomialFunction:
    cdef generic_pd _dag
    cdef object _coeffs

    cdef object _parse_structure(CompiledPolynomialFunction)
    cdef generic_pd _get_gap(CompiledPolynomialFunction, BinaryTree, int)
    cdef void _fill_gaps_binary(CompiledPolynomialFunction, BinaryTree) noexcept
    cdef object eval(CompiledPolynomialFunction, object)

cdef class generic_pd:
    cdef object value
    cdef int refs, hits
    cdef int label
    cdef int eval(self, vars, coeffs) except -2
    cdef generic_pd nodummies(generic_pd)
    cdef void reset(self) noexcept

cdef class dummy_pd(generic_pd):
    cdef generic_pd link
    cdef void fill(dummy_pd self, generic_pd link) noexcept

cdef class var_pd(generic_pd):
    cdef int index

cdef class univar_pd(generic_pd):
    pass

cdef class coeff_pd(generic_pd):
    cdef int index


cdef class unary_pd(generic_pd):
    cdef generic_pd operand

cdef class sqr_pd(unary_pd):
    pass

cdef class pow_pd(unary_pd):
    cdef object exponent


cdef class binary_pd(generic_pd):
    cdef generic_pd left, right

cdef class add_pd(binary_pd):
    pass

cdef class mul_pd(binary_pd):
    pass

cdef class abc_pd(binary_pd):
    cdef int index



# sage_setup: distribution = sagemath-categories
"""
The Spec functor

AUTHORS:

- William Stein (2006): initial implementation

- Peter Bruin (2014): rewrite Spec as a functor
"""

# ******************************************************************************
#  Copyright (C) 2006 William Stein
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ******************************************************************************

from sage.categories.functor import Functor
from sage.rings.integer_ring import ZZ
from sage.schemes.generic.scheme import AffineScheme
from sage.structure.unique_representation import UniqueRepresentation


def Spec(R, S=None):
    r"""
    Apply the Spec functor to `R`.

    INPUT:

    - ``R`` -- either a commutative ring or a ring homomorphism

    - ``S`` -- a commutative ring (optional), the base ring

    OUTPUT:

    - ``AffineScheme`` -- the affine scheme `\mathrm{Spec}(R)`

    EXAMPLES::

        sage: Spec(QQ)
        Spectrum of Rational Field
        sage: Spec(PolynomialRing(QQ, 'x'))
        Spectrum of Univariate Polynomial Ring in x over Rational Field
        sage: Spec(PolynomialRing(QQ, 'x', 3))
        Spectrum of Multivariate Polynomial Ring in x0, x1, x2 over Rational Field
        sage: X = Spec(PolynomialRing(GF(49,'a'), 3, 'x')); X                           # needs sage.rings.finite_rings
        Spectrum of Multivariate Polynomial Ring in x0, x1, x2
         over Finite Field in a of size 7^2
        sage: TestSuite(X).run()                                                        # needs sage.rings.finite_rings

    Applying ``Spec`` twice to the same ring gives identical output
    (see :issue:`17008`)::

        sage: A = Spec(ZZ); B = Spec(ZZ)
        sage: A is B
        True

    A :exc:`TypeError` is raised if the input is not a commutative ring::

        sage: Spec(5)
        Traceback (most recent call last):
        ...
        TypeError: x (=5) is not in Category of commutative rings
        sage: Spec(FreeAlgebra(QQ, 2, 'x'))                                             # needs sage.combinat sage.modules
        Traceback (most recent call last):
        ...
        TypeError: x (=Free Algebra on 2 generators (x0, x1) over Rational Field)
        is not in Category of commutative rings

    TESTS::

        sage: X = Spec(ZZ)
        sage: X
        Spectrum of Integer Ring
        sage: X.base_scheme()
        Spectrum of Integer Ring
        sage: X.base_ring()
        Integer Ring
        sage: X.dimension()
        1
        sage: Spec(QQ, QQ).base_scheme()
        Spectrum of Rational Field
        sage: Spec(RDF, QQ).base_scheme()
        Spectrum of Rational Field
    """
    return SpecFunctor(S)(R)


class SpecFunctor(Functor, UniqueRepresentation):
    """
    The Spec functor.
    """
    def __init__(self, base_ring=None):
        """
        EXAMPLES::

            sage: from sage.schemes.generic.spec import SpecFunctor
            sage: SpecFunctor()
            Spec functor from Category of commutative rings to Category of schemes
            sage: SpecFunctor(QQ)
            Spec functor from Category of commutative rings to
             Category of schemes over Rational Field
        """
        from sage.categories.commutative_rings import CommutativeRings
        from sage.categories.schemes import Schemes

        if base_ring is None:
            domain = CommutativeRings()
            codomain = Schemes()
        elif base_ring in CommutativeRings():
            # We would like to use CommutativeAlgebras(base_ring) as
            # the domain; we use CommutativeRings() instead because
            # currently many algebras are not yet considered to be in
            # CommutativeAlgebras(base_ring) by the category framework.
            domain = CommutativeRings()
            codomain = Schemes(AffineScheme(base_ring))
        else:
            raise TypeError('base (= {}) must be a commutative ring'.format(base_ring))
        self._base_ring = base_ring
        super().__init__(domain, codomain)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.schemes.generic.spec import SpecFunctor
            sage: SpecFunctor(QQ)
            Spec functor from Category of commutative rings to
             Category of schemes over Rational Field
        """
        return 'Spec functor from {} to {}'.format(self.domain(), self.codomain())

    def _latex_(self):
        r"""
        Return a LaTeX representation of ``self``.

        EXAMPLES::

            sage: from sage.schemes.generic.spec import SpecFunctor
            sage: latex(SpecFunctor())
            \mathrm{Spec}\colon \mathbf{CommutativeRings} \longrightarrow \mathbf{Schemes}
        """
        return r'\mathrm{{Spec}}\colon {} \longrightarrow {}'.format(
               self.domain()._latex_(), self.codomain()._latex_())

    def _apply_functor(self, A):
        """
        Apply the Spec functor to the commutative ring ``A``.

        EXAMPLES::

            sage: from sage.schemes.generic.spec import SpecFunctor
            sage: F = SpecFunctor()
            sage: F(RR)  # indirect doctest                                             # needs sage.rings.real_mpfr
            Spectrum of Real Field with 53 bits of precision
        """
        # The second argument of AffineScheme defaults to None.
        # However, AffineScheme has unique representation, so there is
        # a difference between calling it with or without explicitly
        # giving this argument.
        if self._base_ring is None:
            return AffineScheme(A)
        return AffineScheme(A, self._base_ring)

    def _apply_functor_to_morphism(self, f):
        """
        Apply the Spec functor to the ring homomorphism ``f``.

        EXAMPLES::

            sage: from sage.schemes.generic.spec import SpecFunctor
            sage: F = SpecFunctor(GF(7))
            sage: A.<x, y> = GF(7)[]
            sage: B.<t> = GF(7)[]
            sage: f = A.hom((t^2, t^3))
            sage: Spec(f)  # indirect doctest
            Affine Scheme morphism:
              From: Spectrum of Univariate Polynomial Ring in t over Finite Field of size 7
              To:   Spectrum of Multivariate Polynomial Ring in x, y over Finite Field of size 7
              Defn: Ring morphism:
                     From: Multivariate Polynomial Ring in x, y over Finite Field of size 7
                     To:   Univariate Polynomial Ring in t over Finite Field of size 7
                     Defn: x |--> t^2
                           y |--> t^3
        """
        A = f.domain()
        B = f.codomain()
        return self(B).hom(f, self(A))


SpecZ = Spec(ZZ)


# Compatibility with older versions of this module

from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.schemes.generic.spec', 'Spec', AffineScheme)



# sage_setup: distribution = sagemath-categories
from sage.libs.gmp.types cimport *
from sage.rings.finite_rings.stdint cimport *
from sage.rings.finite_rings.element_base cimport FiniteRingElement
from sage.rings.integer cimport Integer


cdef class NativeIntStruct:
    cdef Integer sageInteger
    cdef int_fast32_t int32
    cdef int_fast64_t int64
    cdef readonly list table     # list of elements of IntegerModRing(n)
    cdef readonly list inverses  # list of inverses (None if not invertible)
    cdef inline type element_class(self):
        if self.int32 > 0:
            return IntegerMod_int
        elif self.int64 > 0:
            return IntegerMod_int64
        else:
            return IntegerMod_gmp


cdef class IntegerMod_abstract(FiniteRingElement):
    cdef NativeIntStruct _modulus
    cdef _new_c_from_long(self, long value)
    cdef IntegerMod_abstract _new_c_fast(self, unsigned long value)
    cdef void set_from_mpz(self, mpz_t value) noexcept
    cdef void set_from_long(self, long value) noexcept
    cdef void set_from_ulong_fast(self, unsigned long value) noexcept
    cdef bint is_square_c(self) except -2
    cpdef bint is_one(self) noexcept
    cpdef bint is_unit(self) noexcept
    cpdef _floordiv_(self, other)

cdef class IntegerMod_gmp(IntegerMod_abstract):
    cdef mpz_t value
    cdef IntegerMod_gmp _new_c(self)
    cdef shift(IntegerMod_gmp self, long k)

cdef class IntegerMod_int(IntegerMod_abstract):
    cdef int_fast32_t ivalue
    cdef void set_from_int(IntegerMod_int self, int_fast32_t value) noexcept
    cdef int_fast32_t get_int_value(IntegerMod_int self) noexcept
    cdef IntegerMod_int _new_c(self, int_fast32_t value)
    cdef shift(IntegerMod_int self, int k)

cdef class IntegerMod_int64(IntegerMod_abstract):
    cdef int_fast64_t ivalue
    cdef void set_from_int(IntegerMod_int64 self, int_fast64_t value) noexcept
    cdef int_fast64_t get_int_value(IntegerMod_int64 self) noexcept
    cdef IntegerMod_int64 _new_c(self, int_fast64_t value)
    cdef shift(IntegerMod_int64 self, int k)


cdef int_fast32_t mod_inverse_int(int_fast32_t x, int_fast32_t n) except 0
cdef bint use_32bit_type(int_fast64_t modulus) noexcept



# sage_setup: distribution = sagemath-categories
r"""
LaTeX macros

AUTHORS:

- John H. Palmieri (2009-03)

The code here sets up LaTeX macro definitions for use in the
documentation. To add a macro, modify the list ``macros``, near the
end of this file, and then run 'sage -b'. The entries in this list are
used to produce ``sage_latex_macros``, a list of strings of the form
'\\newcommand...', and ``sage_mathjax_macros``, a list of strings
suitable for parsing by MathJax.  The LaTeX macros are produced using
the ``_latex_`` method for each Sage object listed in ``macros``, and
the MathJax macros are produced from the LaTeX macros.  The list of
LaTeX macros is used in the file
``sage_docbuild.conf`` to add to the preambles of
both the LaTeX file used to build the PDF version of the documentation
and the LaTeX file used to build the HTML version.

Any macro defined here may be used in docstrings or in the tutorial
(or other pieces of documentation).  In a docstring, for example,
"\ZZ" in backquotes (demarking math mode) will appear as "ZZ" in
interactive help, but will be typeset as "\\Bold{Z}" in the
reference manual.

More details on the list ``macros``: the entries are lists or tuples
of the form ``[name]`` or ``[name, arguments]``, where ``name`` is a
string and ``arguments`` consists of valid arguments for the Sage
object named ``name``.  For example, ``["ZZ"]`` and ``["GF", 2]``
produce the LaTeX macros '\\newcommand{\\ZZ}{\\Bold{Z}}' and
'\\newcommand{\\GF}[1]{\\Bold{F}_{#1}}', respectively.  (For the
second of these, ``latex(GF(2))`` is called and the string '2' gets
replaced by '#1', so ``["GF", 17]`` would have worked just as well.
``["GF", p]`` would have raised an error, though, because ``p`` is not
defined, and ``["GF", 4]`` would have raised an error, because to
define the field with four elements in Sage, you also need to specify
the name of a generator.)

To see evidence of the results of the code here, run ``sage --docbuild
tutorial latex`` (for example), and look at the resulting LaTeX file in
``SAGE_DOC/latex/en/tutorial/``.  The preamble should
contain '\newcommand' lines for each of the entries in ``macros``.
"""


def produce_latex_macro(name, *sample_args):
    r"""
    Produce a string defining a LaTeX macro.

    INPUT:

    - ``name`` -- name of macro to be defined, also name of corresponding Sage object

    - ``sample_args`` -- (optional) sample arguments for this Sage object

    EXAMPLES::

        sage: from sage.misc.latex_macros import produce_latex_macro
        sage: produce_latex_macro('ZZ')
        '\\newcommand{\\ZZ}{\\Bold{Z}}'

    If the Sage object takes arguments, then the LaTeX macro will
    accept arguments as well. You must pass valid arguments, which
    will then be converted to #1, #2, etc. in the macro
    definition. The following allows the use of "\GF{p^n}", for
    example::

         sage: produce_latex_macro('GF', 37)
         '\\newcommand{\\GF}[1]{\\Bold{F}_{#1}}'

    If the Sage object is not in the global name space, describe it
    like so::

         sage: produce_latex_macro('sage.rings.finite_rings.finite_field_constructor.FiniteField', 3)
         '\\newcommand{\\FiniteField}[1]{\\Bold{F}_{#1}}'
    """
    from sage.misc.latex import LatexCall  # type: ignore
    # this import is used inside a string below
    names_split = name.rsplit('.', 1)
    if len(names_split) == 1:

        try:
            import sage.all as toplevel
        except ImportError:
            try:
                import sage.all__sagemath_polyhedra as toplevel
            except ImportError:
                try:
                    import sage.all__sagemath_modules as toplevel
                except ImportError:
                    try:
                        import sage.all__sagemath_categories as toplevel
                    except ImportError:
                        import sage.all__sagemath_objects as toplevel
        module = toplevel.__name__
        real_name = names_split[0]
    else:
        module, real_name = names_split
    newcommand = '\\newcommand{\\' + real_name + '}'
    count = 0
    args = "("
    for x in sample_args:
        count += 1
        args += str(x) + ','
    args += ')'
    exec('from ' + module + ' import ' + real_name)
    if count:
        defn = '[' + str(count) + ']{'
        defn += eval('str(LatexCall()(' + real_name + args + '))') + '}'
    else:
        defn = '{' + eval('str(LatexCall()(' + real_name + '))') + '}'
    count = 0
    for x in sample_args:
        count += 1
        defn = defn.replace(str(x), "#" + str(count))
    return newcommand + defn


def convert_latex_macro_to_mathjax(macro):
    r"""
    This converts a LaTeX macro definition (\newcommand...) to a
    MathJax macro definition (MathJax.Macro...).

    INPUT:

    - ``macro`` -- LaTeX macro definition

    See the web page
    https://docs.mathjax.org/en/latest/input/tex/macros.html for a
    description of the format for MathJax macros.

    EXAMPLES::

        sage: from sage.misc.latex_macros import convert_latex_macro_to_mathjax
        sage: convert_latex_macro_to_mathjax('\\newcommand{\\ZZ}{\\Bold{Z}}')
        ('ZZ', '\\Bold{Z}')
        sage: convert_latex_macro_to_mathjax('\\newcommand{\\GF}[1]{\\Bold{F}_{#1}}')
        ('GF', ['\\Bold{F}_{#1}', 1])
    """
    left_bracket = macro.find('[')
    right_bracket = macro.find('[')
    if left_bracket >= 0:
        right_bracket = macro.find(']')
        num_args = int(macro[left_bracket + 1: right_bracket])
    else:
        num_args = 0
    start_name = macro.find('{') + 1  # add one to go past the backslash
    end_name = macro.find('}')
    name = macro[start_name + 1: end_name]
    start_defn = macro.find('{', end_name)
    end_defn = macro.rfind('}')
    defn = macro[start_defn + 1: end_defn]
    if num_args == 0:
        return name, defn
    else:
        return name, [defn, num_args]


# To add a new macro for use in the Sage documentation, add a list or
# tuple to the following list.  Each list (or tuple) should have the
# form [name, arguments], which will be passed to the function
# produce_latex_macro: see that for more documentation.
#
# To see the results of this, run 'sage --docbuild tutorial latex' (for
# example -- you could replace 'tutorial' with your favorite piece of
# documentation), and look at the resulting tex file in
# SAGE_DOC/latex/en/tutorial.  The preamble should contain
# \newcommand's for each of the entries here.
macros = [["ZZ"],
          ["NN"],
          ["RR"],
          ["CC"],
          ["QQ"],
          ["QQbar"],
          ["GF", 2],
          ["Zp", 2],
          ["Qp", 2],
          ["Zmod", 2],
          ["CDF"],
          ["CIF"],
          ["CLF"],
          ["RDF"],
          ["RIF"],
          ["RLF"],
          ]

# Use this list to define additional latex macros for sage documentation
latex_macros = [r"\newcommand{\SL}{\mathrm{SL}}",
                r"\newcommand{\PSL}{\mathrm{PSL}}",
                r"\newcommand{\lcm}{\mathop{\operatorname{lcm}}}",
                r"\newcommand{\dist}{\mathrm{dist}}"]

# The following is to allow customization of typesetting of rings:
# mathbf vs mathbb.  See latex.py for more information.
sage_configurable_latex_macros = [r"\newcommand{\Bold}[1]{\mathbf{#1}}"]


def sage_latex_macros():
    r"""
    Return list of LaTeX macros for Sage. This just runs the function
    :func:`produce_latex_macro` on the list ``macros`` defined in this
    file, and appends ``sage_configurable_latex_macros``. To add a new
    macro for permanent use in Sage, modify ``macros``.

    EXAMPLES::

        sage: from sage.misc.latex_macros import sage_latex_macros
        sage: sage_latex_macros()
        ['\\newcommand{\\ZZ}{\\Bold{Z}}', '\\newcommand{\\NN}{\\Bold{N}}', ...
    """
    result = []
    for x in macros:
        try:
            result.append(produce_latex_macro(*x))
        except ImportError:
            pass
    return result + latex_macros + sage_configurable_latex_macros


def sage_mathjax_macros():
    r"""
    Return Sage's macro definitions for usage with MathJax.

    This feeds each item output by :func:`sage_latex_macros` to
    :func:`convert_latex_macro_to_mathjax`.

    EXAMPLES::

        sage: from sage.misc.latex_macros import sage_mathjax_macros
        sage: sage_mathjax_macros()
        {'Bold': ['\\mathbf{#1}', 1], ... 'QQ': '\\Bold{Q}', ...
    """
    return dict(convert_latex_macro_to_mathjax(m) for m in sage_latex_macros())



# sage_setup: distribution = sagemath-categories
r"""
Number field elements (abstract base class)
"""

# ****************************************************************************
#       Copyright (C) 2023 Matthias Koeppe
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

cdef class NumberFieldElement_base(FieldElement):
    r"""
    Abstract base class for
    :class:`~sage.rings.number_field.number_field_element.NumberFieldElement`.

    This class is defined for the purpose of :func:`isinstance` tests.
    It should not be instantiated.

    EXAMPLES::

        sage: x = polygen(ZZ, 'x')
        sage: k.<a> = NumberField(x^3 + x + 1)                                          # needs sage.rings.number_field
        sage: isinstance(a, sage.rings.number_field.number_field_element_base.NumberFieldElement_base)                  # needs sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: len(sage.rings.number_field.number_field_element_base.NumberFieldElement_base.__subclasses__()) <= 1
        True
    """

    pass
